@@ -104,7 +104,8 @@ END(BorderHashKey)
 START(filter, filter_desc)
     GENERIC(int, flags)
     GENERIC(int, be)
-    GENERIC(int, blur)
+    GENERIC(int, blur_x)
+    GENERIC(int, blur_y)
     VECTOR(shadow)
 END(FilterDesc)
 
