@@ -49,10 +49,9 @@ struct ass_font {
     double size;
 };
 
-void charmap_magic(ASS_Library *library, FT_Face face);
+void ass_charmap_magic(ASS_Library *library, FT_Face face);
 ASS_Font *ass_font_new(ASS_Renderer *render_priv, ASS_FontDesc *desc);
 void ass_face_set_size(FT_Face face, double size);
-void ass_font_set_size(ASS_Font *font, double size);
 int ass_face_get_weight(FT_Face face);
 void ass_font_get_asc_desc(ASS_Font *font, int face_index,
                            int *asc, int *desc);
