@@ -17,6 +17,7 @@
  */
 
 #include "config.h"
+#include "ass_compat.h"
 
 #include "ass_shaper.h"
 #include "ass_render.h"
@@ -24,9 +25,11 @@
 #include "ass_parse.h"
 #include "ass_cache.h"
 #include <limits.h>
+#include <stdbool.h>
 
-#ifdef CONFIG_HARFBUZZ
-#include <hb-ft.h>
+#include <ft2build.h>
+#include FT_FREETYPE_H
+#include FT_TRUETYPE_TABLES_H
 enum {
     VERT = 0,
     VKNA,
@@ -35,20 +38,25 @@ enum {
     CLIG
 };
 #define NUM_FEATURES 5
-#endif
+
+enum {
+    WHOLE_TEXT_LAYOUT_OFF,
+    WHOLE_TEXT_LAYOUT_IMPLICIT,
+    WHOLE_TEXT_LAYOUT_EXPLICIT,
+};
 
 struct ass_shaper {
     ASS_ShapingLevel shaping_level;
 
     // FriBidi log2vis
-    int n_glyphs;
-    FriBidiChar *event_text;
+    int n_glyphs, n_pars;
+    FriBidiChar *event_text; // just a reference, owned by text_info
     FriBidiCharType *ctypes;
     FriBidiLevel *emblevels;
     FriBidiStrIndex *cmap;
+    FriBidiParType *pbase_dir;
     FriBidiParType base_direction;
 
-#ifdef CONFIG_HARFBUZZ
     // OpenType features
     int n_features;
     hb_feature_t *features;
@@ -56,10 +64,15 @@ struct ass_shaper {
 
     // Glyph metrics cache, to speed up shaping
     Cache *metrics_cache;
+
+#ifdef USE_FRIBIDI_EX_API
+    FriBidiBracketType *btypes;
+    bool bidi_brackets;
 #endif
+
+    char whole_text_layout;
 };
 
-#ifdef CONFIG_HARFBUZZ
 struct ass_shaper_metrics_data {
     Cache *metrics_cache;
     GlyphMetricsHashKey hash_key;
@@ -71,18 +84,15 @@ struct ass_shaper_font_data {
     hb_font_funcs_t *font_funcs[ASS_FONT_MAX_FACES];
     struct ass_shaper_metrics_data *metrics_data[ASS_FONT_MAX_FACES];
 };
-#endif
 
 /**
  * \brief Print version information
  */
 void ass_shaper_info(ASS_Library *lib)
 {
-    ass_msg(lib, MSGL_V, "Shaper: FriBidi "
+    ass_msg(lib, MSGL_INFO, "Shaper: FriBidi "
             FRIBIDI_VERSION " (SIMPLE)"
-#ifdef CONFIG_HARFBUZZ
             " HarfBuzz-ng %s (COMPLEX)", hb_version_string()
-#endif
            );
 }
 
@@ -90,14 +100,24 @@ void ass_shaper_info(ASS_Library *lib)
  * \brief grow arrays, if needed
  * \param new_size requested size
  */
-static void check_allocations(ASS_Shaper *shaper, size_t new_size)
+static bool check_allocations(ASS_Shaper *shaper, size_t new_size, size_t n_pars)
 {
     if (new_size > shaper->n_glyphs) {
-        shaper->event_text = realloc(shaper->event_text, sizeof(FriBidiChar) * new_size);
-        shaper->ctypes     = realloc(shaper->ctypes, sizeof(FriBidiCharType) * new_size);
-        shaper->emblevels  = realloc(shaper->emblevels, sizeof(FriBidiLevel) * new_size);
-        shaper->cmap       = realloc(shaper->cmap, sizeof(FriBidiStrIndex) * new_size);
+        if (!ASS_REALLOC_ARRAY(shaper->ctypes, new_size) ||
+#ifdef USE_FRIBIDI_EX_API
+            (shaper->bidi_brackets && !ASS_REALLOC_ARRAY(shaper->btypes, new_size)) ||
+#endif
+            !ASS_REALLOC_ARRAY(shaper->emblevels, new_size) ||
+            !ASS_REALLOC_ARRAY(shaper->cmap, new_size))
+            return false;
+        shaper->n_glyphs = new_size;
     }
+    if (shaper->whole_text_layout && n_pars > shaper->n_pars) {
+        if (!ASS_REALLOC_ARRAY(shaper->pbase_dir, n_pars))
+            return false;
+        shaper->n_pars = n_pars;
+    }
+    return true;
 }
 
 /**
@@ -105,20 +125,19 @@ static void check_allocations(ASS_Shaper *shaper, size_t new_size)
  */
 void ass_shaper_free(ASS_Shaper *shaper)
 {
-#ifdef CONFIG_HARFBUZZ
-    ass_cache_done(shaper->metrics_cache);
     free(shaper->features);
-#endif
-    free(shaper->event_text);
     free(shaper->ctypes);
+#ifdef USE_FRIBIDI_EX_API
+    free(shaper->btypes);
+#endif
     free(shaper->emblevels);
     free(shaper->cmap);
+    free(shaper->pbase_dir);
     free(shaper);
 }
 
 void ass_shaper_font_data_free(ASS_ShaperFontData *priv)
 {
-#ifdef CONFIG_HARFBUZZ
     int i;
     for (i = 0; i < ASS_FONT_MAX_FACES; i++)
         if (priv->fonts[i]) {
@@ -127,17 +146,17 @@ void ass_shaper_font_data_free(ASS_ShaperFontData *priv)
             hb_font_funcs_destroy(priv->font_funcs[i]);
         }
     free(priv);
-#endif
 }
 
-#ifdef CONFIG_HARFBUZZ
 /**
  * \brief set up the HarfBuzz OpenType feature list with some
  * standard features.
  */
-static void init_features(ASS_Shaper *shaper)
+static bool init_features(ASS_Shaper *shaper)
 {
     shaper->features = calloc(sizeof(hb_feature_t), NUM_FEATURES);
+    if (!shaper->features)
+        return false;
 
     shaper->n_features = NUM_FEATURES;
     shaper->features[VERT].tag = HB_TAG('v', 'e', 'r', 't');
@@ -150,6 +169,8 @@ static void init_features(ASS_Shaper *shaper)
     shaper->features[LIGA].end = UINT_MAX;
     shaper->features[CLIG].tag = HB_TAG('c', 'l', 'i', 'g');
     shaper->features[CLIG].end = UINT_MAX;
+
+    return true;
 }
 
 /**
@@ -194,106 +215,160 @@ static void update_hb_size(hb_font_t *hb_font, FT_Face face)
  *
  */
 
-GlyphMetricsHashValue *
-get_cached_metrics(struct ass_shaper_metrics_data *metrics, FT_Face face,
+static FT_Glyph_Metrics *
+get_cached_metrics(struct ass_shaper_metrics_data *metrics,
                    hb_codepoint_t unicode, hb_codepoint_t glyph)
 {
-    GlyphMetricsHashValue *val;
+    bool rotate = false;
+    // if @font rendering is enabled and the glyph should be rotated,
+    // make cached_h_advance pick up the right advance later
+    if (metrics->vertical && unicode >= VERTICAL_LOWER_BOUND)
+        rotate = true;
 
     metrics->hash_key.glyph_index = glyph;
-    val = ass_cache_get(metrics->metrics_cache, &metrics->hash_key);
+    FT_Glyph_Metrics *val = ass_cache_get(metrics->metrics_cache, &metrics->hash_key,
+                                          rotate ? metrics : NULL);
+    if (!val)
+        return NULL;
+    if (val->width >= 0)
+        return val;
+    ass_cache_dec_ref(val);
+    return NULL;
+}
 
-    if (!val) {
-        int load_flags = FT_LOAD_DEFAULT | FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH
-            | FT_LOAD_IGNORE_TRANSFORM;
-        GlyphMetricsHashValue new_val;
+size_t ass_glyph_metrics_construct(void *key, void *value, void *priv)
+{
+    GlyphMetricsHashKey *k = key;
+    FT_Glyph_Metrics *v = value;
 
-        if (FT_Load_Glyph(face, glyph, load_flags))
-            return NULL;
+    int load_flags = FT_LOAD_DEFAULT | FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH
+        | FT_LOAD_IGNORE_TRANSFORM;
 
-        memcpy(&new_val.metrics, &face->glyph->metrics, sizeof(FT_Glyph_Metrics));
-
-        // if @font rendering is enabled and the glyph should be rotated,
-        // make cached_h_advance pick up the right advance later
-        if (metrics->vertical && unicode >= VERTICAL_LOWER_BOUND)
-            new_val.metrics.horiAdvance = new_val.metrics.vertAdvance;
-
-        val = ass_cache_put(metrics->metrics_cache, &metrics->hash_key, &new_val);
+    FT_Face face = k->font->faces[k->face_index];
+    if (FT_Load_Glyph(face, k->glyph_index, load_flags)) {
+        v->width = -1;
+        return 1;
     }
 
-    return val;
+    memcpy(v, &face->glyph->metrics, sizeof(FT_Glyph_Metrics));
+
+    if (priv)  // rotate
+        v->horiAdvance = v->vertAdvance;
+
+    return 1;
+}
+
+static hb_blob_t*
+get_reference_table(hb_face_t *hbface, hb_tag_t tag, void *font_data)
+{
+  FT_Face face = font_data;
+  FT_ULong len = 0;
+
+  if (FT_Load_Sfnt_Table(face, tag, 0, NULL, &len) != FT_Err_Ok)
+    return NULL;
+
+  char *buf = malloc(len);
+  if (!buf)
+    return NULL;
+
+  if (FT_Load_Sfnt_Table(face, tag, 0, (FT_Byte*)buf, &len) != FT_Err_Ok) {
+    free(buf);
+    return NULL;
+  }
+
+  hb_blob_t *blob = hb_blob_create(buf, len, HB_MEMORY_MODE_WRITABLE, buf, free);
+  if (!blob)
+      free(buf);
+
+  return blob;
 }
 
 static hb_bool_t
-get_glyph(hb_font_t *font, void *font_data, hb_codepoint_t unicode,
-          hb_codepoint_t variation, hb_codepoint_t *glyph, void *user_data)
+get_glyph_nominal(hb_font_t *font, void *font_data, hb_codepoint_t unicode,
+                  hb_codepoint_t *glyph, void *user_data)
 {
     FT_Face face = font_data;
     struct ass_shaper_metrics_data *metrics_priv = user_data;
 
-    if (variation)
-        *glyph = FT_Face_GetCharVariantIndex(face, ass_font_index_magic(face, unicode), variation);
-    else
-        *glyph = FT_Get_Char_Index(face, ass_font_index_magic(face, unicode));
+    *glyph = ass_font_index_magic(face, unicode);
+    if (*glyph)
+        *glyph = FT_Get_Char_Index(face, *glyph);
+    if (!*glyph)
+        return false;
 
     // rotate glyph advances for @fonts while we still know the Unicode codepoints
-    if (*glyph != 0)
-        get_cached_metrics(metrics_priv, face, unicode, *glyph);
+    FT_Glyph_Metrics *metrics = get_cached_metrics(metrics_priv, unicode, *glyph);
+    ass_cache_dec_ref(metrics);
+    return true;
+}
 
-    return *glyph != 0;
+static hb_bool_t
+get_glyph_variation(hb_font_t *font, void *font_data, hb_codepoint_t unicode,
+                    hb_codepoint_t variation, hb_codepoint_t *glyph, void *user_data)
+{
+    FT_Face face = font_data;
+    struct ass_shaper_metrics_data *metrics_priv = user_data;
+
+    *glyph = ass_font_index_magic(face, unicode);
+    if (*glyph)
+        *glyph = FT_Face_GetCharVariantIndex(face, *glyph, variation);
+    if (!*glyph)
+        return false;
+
+    // rotate glyph advances for @fonts while we still know the Unicode codepoints
+    FT_Glyph_Metrics *metrics = get_cached_metrics(metrics_priv, unicode, *glyph);
+    ass_cache_dec_ref(metrics);
+    return true;
 }
 
 static hb_position_t
 cached_h_advance(hb_font_t *font, void *font_data, hb_codepoint_t glyph,
                  void *user_data)
 {
-    FT_Face face = font_data;
     struct ass_shaper_metrics_data *metrics_priv = user_data;
-    GlyphMetricsHashValue *metrics = get_cached_metrics(metrics_priv, face, 0, glyph);
-
+    FT_Glyph_Metrics *metrics = get_cached_metrics(metrics_priv, 0, glyph);
     if (!metrics)
         return 0;
 
-    return metrics->metrics.horiAdvance;
+    hb_position_t advance = metrics->horiAdvance;
+    ass_cache_dec_ref(metrics);
+    return advance;
 }
 
 static hb_position_t
 cached_v_advance(hb_font_t *font, void *font_data, hb_codepoint_t glyph,
                  void *user_data)
 {
-    FT_Face face = font_data;
     struct ass_shaper_metrics_data *metrics_priv = user_data;
-    GlyphMetricsHashValue *metrics = get_cached_metrics(metrics_priv, face, 0, glyph);
-
+    FT_Glyph_Metrics *metrics = get_cached_metrics(metrics_priv, 0, glyph);
     if (!metrics)
         return 0;
 
-    return metrics->metrics.vertAdvance;
-
+    hb_position_t advance = metrics->vertAdvance;
+    ass_cache_dec_ref(metrics);
+    return advance;
 }
 
 static hb_bool_t
 cached_h_origin(hb_font_t *font, void *font_data, hb_codepoint_t glyph,
                 hb_position_t *x, hb_position_t *y, void *user_data)
 {
-    return 1;
+    return true;
 }
 
 static hb_bool_t
 cached_v_origin(hb_font_t *font, void *font_data, hb_codepoint_t glyph,
                 hb_position_t *x, hb_position_t *y, void *user_data)
 {
-    FT_Face face = font_data;
     struct ass_shaper_metrics_data *metrics_priv = user_data;
-    GlyphMetricsHashValue *metrics = get_cached_metrics(metrics_priv, face, 0, glyph);
-
+    FT_Glyph_Metrics *metrics = get_cached_metrics(metrics_priv, 0, glyph);
     if (!metrics)
-        return 0;
+        return false;
 
-    *x = metrics->metrics.horiBearingX - metrics->metrics.vertBearingX;
-    *y = metrics->metrics.horiBearingY - (-metrics->metrics.vertBearingY);
-
-    return 1;
+    *x = metrics->horiBearingX - metrics->vertBearingX;
+    *y = metrics->horiBearingY + metrics->vertBearingY;
+    ass_cache_dec_ref(metrics);
+    return true;
 }
 
 static hb_position_t
@@ -303,7 +378,7 @@ get_h_kerning(hb_font_t *font, void *font_data, hb_codepoint_t first,
     FT_Face face = font_data;
     FT_Vector kern;
 
-    if (FT_Get_Kerning (face, first, second, FT_KERNING_DEFAULT, &kern))
+    if (FT_Get_Kerning(face, first, second, FT_KERNING_DEFAULT, &kern))
         return 0;
 
     return kern.x;
@@ -320,19 +395,17 @@ static hb_bool_t
 cached_extents(hb_font_t *font, void *font_data, hb_codepoint_t glyph,
                hb_glyph_extents_t *extents, void *user_data)
 {
-    FT_Face face = font_data;
     struct ass_shaper_metrics_data *metrics_priv = user_data;
-    GlyphMetricsHashValue *metrics = get_cached_metrics(metrics_priv, face, 0, glyph);
-
+    FT_Glyph_Metrics *metrics = get_cached_metrics(metrics_priv, 0, glyph);
     if (!metrics)
-        return 0;
+        return false;
 
-    extents->x_bearing = metrics->metrics.horiBearingX;
-    extents->y_bearing = metrics->metrics.horiBearingY;
-    extents->width     = metrics->metrics.width;
-    extents->height    = -metrics->metrics.height;
-
-    return 1;
+    extents->x_bearing =  metrics->horiBearingX;
+    extents->y_bearing =  metrics->horiBearingY;
+    extents->width     =  metrics->width;
+    extents->height    = -metrics->height;
+    ass_cache_dec_ref(metrics);
+    return true;
 }
 
 static hb_bool_t
@@ -345,15 +418,14 @@ get_contour_point(hb_font_t *font, void *font_data, hb_codepoint_t glyph,
         | FT_LOAD_IGNORE_TRANSFORM;
 
     if (FT_Load_Glyph(face, glyph, load_flags))
-        return 0;
+        return false;
 
     if (point_index >= (unsigned)face->glyph->outline.n_points)
-        return 0;
+        return false;
 
     *x = face->glyph->outline.points[point_index].x;
     *y = face->glyph->outline.points[point_index].y;
-
-    return 1;
+    return true;
 }
 
 /**
@@ -369,24 +441,43 @@ static hb_font_t *get_hb_font(ASS_Shaper *shaper, GlyphInfo *info)
 
     if (!font->shaper_priv)
         font->shaper_priv = calloc(sizeof(ASS_ShaperFontData), 1);
-
+    if (!font->shaper_priv)
+        return NULL;
 
     hb_fonts = font->shaper_priv->fonts;
     if (!hb_fonts[info->face_index]) {
-        hb_fonts[info->face_index] =
-            hb_ft_font_create(font->faces[info->face_index], NULL);
+        FT_Face face = font->faces[info->face_index];
+        hb_face_t *hb_face = hb_face_create_for_tables(get_reference_table, face, NULL);
+        if (!hb_face)
+            return NULL;
+        hb_face_set_index(hb_face, face->face_index);
+        hb_face_set_upem(hb_face, face->units_per_EM);
+
+        hb_font_t *hb_font = hb_fonts[info->face_index] = hb_font_create(hb_face);
+        hb_face_destroy(hb_face);
+        if (!hb_font)
+            return NULL;
+
+        hb_font_set_scale(hb_font,
+            (int)(((uint64_t)face->size->metrics.x_scale * face->units_per_EM + (1<<15)) >> 16),
+            (int)(((uint64_t)face->size->metrics.y_scale * face->units_per_EM + (1<<15)) >> 16));
 
         // set up cached metrics access
-        font->shaper_priv->metrics_data[info->face_index] =
-            calloc(sizeof(struct ass_shaper_metrics_data), 1);
         struct ass_shaper_metrics_data *metrics =
-            font->shaper_priv->metrics_data[info->face_index];
+            font->shaper_priv->metrics_data[info->face_index] =
+                calloc(sizeof(struct ass_shaper_metrics_data), 1);
+        if (!metrics)
+            return NULL;
         metrics->metrics_cache = shaper->metrics_cache;
         metrics->vertical = info->font->desc.vertical;
 
         hb_font_funcs_t *funcs = hb_font_funcs_create();
+        if (!funcs)
+            return NULL;
         font->shaper_priv->font_funcs[info->face_index] = funcs;
-        hb_font_funcs_set_glyph_func(funcs, get_glyph,
+        hb_font_funcs_set_nominal_glyph_func(funcs, get_glyph_nominal,
+                metrics, NULL);
+        hb_font_funcs_set_variation_glyph_func(funcs, get_glyph_variation,
                 metrics, NULL);
         hb_font_funcs_set_glyph_h_advance_func(funcs, cached_h_advance,
                 metrics, NULL);
@@ -404,8 +495,7 @@ static hb_font_t *get_hb_font(ASS_Shaper *shaper, GlyphInfo *info)
                 metrics, NULL);
         hb_font_funcs_set_glyph_contour_point_func(funcs, get_contour_point,
                 metrics, NULL);
-        hb_font_set_funcs(hb_fonts[info->face_index], funcs,
-                font->faces[info->face_index], NULL);
+        hb_font_set_funcs(hb_font, funcs, face, NULL);
     }
 
     ass_face_set_size(font->faces[info->face_index], info->font_size);
@@ -417,12 +507,19 @@ static hb_font_t *get_hb_font(ASS_Shaper *shaper, GlyphInfo *info)
     metrics->hash_key.font = info->font;
     metrics->hash_key.face_index = info->face_index;
     metrics->hash_key.size = info->font_size;
-    metrics->hash_key.scale_x = double_to_d6(info->scale_x);
-    metrics->hash_key.scale_y = double_to_d6(info->scale_y);
 
     return hb_fonts[info->face_index];
 }
 
+/**
+ * \brief Determine whether this Unicode codepoint affects shaping
+ * of neighbors even if they are in separate shape runs due to bidi,
+ * script or font splitting, using VSFilter as the reference.
+ */
+static inline bool is_shaping_control(unsigned symbol) {
+    return symbol == 0x200C /* ZWNJ */ || symbol == 0x200D /* ZWJ */;
+}
+
 /**
  * \brief Map script to default language.
  *
@@ -546,22 +643,25 @@ shape_harfbuzz_process_run(GlyphInfo *glyphs, hb_buffer_t *buf, int offset)
 
         // if we have more than one glyph per cluster, allocate a new one
         // and attach to the root glyph
-        if (info->skip == 0) {
+        if (!info->skip) {
             while (info->next)
                 info = info->next;
             info->next = malloc(sizeof(GlyphInfo));
-            memcpy(info->next, info, sizeof(GlyphInfo));
-            info = info->next;
-            info->next = NULL;
+            if (info->next) {
+                memcpy(info->next, info, sizeof(GlyphInfo));
+                ass_cache_inc_ref(info->font);
+                info = info->next;
+                info->next = NULL;
+            }
         }
 
         // set position and advance
-        info->skip = 0;
+        info->skip = false;
         info->glyph_index = glyph_info[j].codepoint;
-        info->offset.x    = pos[j].x_offset * info->scale_x;
-        info->offset.y    = -pos[j].y_offset * info->scale_y;
-        info->advance.x   = pos[j].x_advance * info->scale_x;
-        info->advance.y   = -pos[j].y_advance * info->scale_y;
+        info->offset.x    = lrint(pos[j].x_offset * info->scale_x);
+        info->offset.y    = lrint(-pos[j].y_offset * info->scale_y);
+        info->advance.x   = lrint(pos[j].x_advance * info->scale_x);
+        info->advance.y   = lrint(-pos[j].y_advance * info->scale_y);
 
         // accumulate advance in the root glyph
         root->cluster_advance.x += info->advance.x;
@@ -574,7 +674,7 @@ shape_harfbuzz_process_run(GlyphInfo *glyphs, hb_buffer_t *buf, int offset)
  * \param glyphs glyph clusters
  * \param len number of clusters
  */
-static void shape_harfbuzz(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
+static bool shape_harfbuzz(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
 {
     int i;
     hb_buffer_t *buf = hb_buffer_create();
@@ -582,23 +682,48 @@ static void shape_harfbuzz(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
 
     // Initialize: skip all glyphs, this is undone later as needed
     for (i = 0; i < len; i++)
-        glyphs[i].skip = 1;
+        glyphs[i].skip = true;
 
     for (i = 0; i < len; i++) {
+        if (glyphs[i].drawing_text.str) {
+            glyphs[i].skip = false;
+            continue;
+        }
+
         int offset = i;
         hb_font_t *font = get_hb_font(shaper, glyphs + offset);
-        int level = glyphs[offset].shape_run_id;
-        int direction = shaper->emblevels[offset] % 2;
+        if (!font)
+            return false;
+        int run_id = glyphs[offset].shape_run_id;
+        int level = shaper->emblevels[offset];
 
         // advance in text until end of run
-        while (i < (len - 1) && level == glyphs[i+1].shape_run_id)
+        while (i < (len - 1) && run_id == glyphs[i + 1].shape_run_id &&
+                level == shaper->emblevels[i + 1])
             i++;
 
         hb_buffer_pre_allocate(buf, i - offset + 1);
-        hb_buffer_add_utf32(buf, shaper->event_text + offset, i - offset + 1,
-                0, i - offset + 1);
 
-        props.direction = direction ? HB_DIRECTION_RTL : HB_DIRECTION_LTR;
+        int lead_context = 0, trail_context = 0;
+        if (shaper->whole_text_layout) {
+            hb_buffer_add_utf32(buf, shaper->event_text, len,
+                    offset, i - offset + 1);
+        } else {
+            if (offset > 0 && !glyphs[offset].starts_new_run &&
+                    is_shaping_control(glyphs[offset - 1].symbol))
+                lead_context = 1;
+            if (i < (len - 1) && !glyphs[i + 1].starts_new_run &&
+                    is_shaping_control(glyphs[i + 1].symbol))
+                trail_context = 1;
+
+            hb_buffer_add_utf32(buf,
+                    shaper->event_text + offset - lead_context,
+                    i - offset + 1 + lead_context + trail_context,
+                    lead_context, i - offset + 1);
+        }
+
+        props.direction = FRIBIDI_LEVEL_IS_RTL(level) ?
+            HB_DIRECTION_RTL : HB_DIRECTION_LTR;
         props.script = glyphs[offset].script;
         props.language  = hb_shaper_get_run_language(shaper, props.script);
         hb_buffer_set_segment_properties(buf, &props);
@@ -606,11 +731,14 @@ static void shape_harfbuzz(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
         set_run_features(shaper, glyphs + offset);
         hb_shape(font, buf, shaper->features, shaper->n_features);
 
-        shape_harfbuzz_process_run(glyphs, buf, offset);
+        shape_harfbuzz_process_run(glyphs, buf,
+                shaper->whole_text_layout ? 0 : offset - lead_context);
         hb_buffer_reset(buf);
     }
 
     hb_buffer_destroy(buf);
+
+    return true;
 }
 
 /**
@@ -662,7 +790,6 @@ void ass_shaper_determine_script(ASS_Shaper *shaper, GlyphInfo *glyphs,
         }
     }
 }
-#endif
 
 /**
  * \brief Shape event text with FriBidi. Does mirroring and simple
@@ -685,7 +812,9 @@ static void shape_fribidi(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
         GlyphInfo *info = glyphs + i;
         FT_Face face = info->font->faces[info->face_index];
         info->symbol = shaper->event_text[i];
-        info->glyph_index = FT_Get_Char_Index(face, ass_font_index_magic(face, shaper->event_text[i]));
+        info->glyph_index = ass_font_index_magic(face, shaper->event_text[i]);
+        if (info->glyph_index)
+            info->glyph_index = FT_Get_Char_Index(face, info->glyph_index);
     }
 
     free(joins);
@@ -693,14 +822,48 @@ static void shape_fribidi(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
 
 /**
  * \brief Toggle kerning for HarfBuzz shaping.
- * NOTE: currently only works with OpenType fonts, the TrueType fallback *always*
- * kerns. It's a bug in HarfBuzz.
+ * \param shaper shaper instance
+ * \param kern toggle kerning
  */
-void ass_shaper_set_kerning(ASS_Shaper *shaper, int kern)
+void ass_shaper_set_kerning(ASS_Shaper *shaper, bool kern)
 {
-#ifdef CONFIG_HARFBUZZ
-    shaper->features[KERN].value = !!kern;
-#endif
+    shaper->features[KERN].value = kern;
+}
+
+/**
+ * \brief Determine whether HarfBuzz ignores any font-provided glyph
+ * for this Unicode codepoint and replaces it with a zero-width glyph.
+ * Matches hb_unicode_funcs_t::is_default_ignorable in hb-unicode.hh.
+ * The affected codepoints are a subset of Unicode's Default_Ignorable list.
+ */
+static inline bool is_harfbuzz_ignorable(unsigned symbol) {
+    switch (symbol >> 8) {
+        case 0x00: return symbol == 0x00AD;
+        case 0x03: return symbol == 0x034F;
+        case 0x06: return symbol == 0x061C;
+        case 0x17: return symbol >= 0x17B4 && symbol <= 0x17B5;
+        case 0x18: return symbol >= 0x180B && symbol <= 0x180E;
+        case 0x20: return (symbol >= 0x200B && symbol <= 0x200F) ||
+                          (symbol >= 0x202A && symbol <= 0x202E) ||
+                          (symbol >= 0x2060 && symbol <= 0x206F);
+        case 0xFE: return (symbol >= 0xFE00 && symbol <= 0xFE0F) ||
+                          symbol == 0xFEFF;
+        case 0xFF: return symbol >= 0xFFF0 && symbol <= 0xFFF8;
+        case 0x1D1: return symbol >= 0x1D173 && symbol <= 0x1D17A;
+        default: return symbol >= 0xE0000 && symbol <= 0xE0FFF;
+    }
+}
+
+/**
+  * \brief Remove all zero-width invisible characters from the text.
+  */
+static void ass_shaper_skip_characters(GlyphInfo *glyphs, size_t len)
+{
+    for (int i = 0; i < len; i++) {
+        if (is_harfbuzz_ignorable(glyphs[i].symbol)) {
+            glyphs[i].skip = true;
+        }
+    }
 }
 
 /**
@@ -712,52 +875,30 @@ void ass_shaper_find_runs(ASS_Shaper *shaper, ASS_Renderer *render_priv,
     int i;
     int shape_run = 0;
 
-#ifdef CONFIG_HARFBUZZ
     ass_shaper_determine_script(shaper, glyphs, len);
-#endif
+    ass_shaper_skip_characters(glyphs, len);
 
     // find appropriate fonts for the shape runs
     for (i = 0; i < len; i++) {
-        GlyphInfo *last = glyphs + i - 1;
         GlyphInfo *info = glyphs + i;
-        // skip drawings
-        if (info->symbol == 0xfffc)
-            continue;
-        // set size and get glyph index
-        ass_font_get_index(render_priv->fontconfig_priv, info->font,
-                info->symbol, &info->face_index, &info->glyph_index);
-        // shape runs break on: xbord, ybord, xshad, yshad,
-        // all four colors, all four alphas, be, blur, fn, fs,
-        // fscx, fscy, fsp, bold, italic, underline, strikeout,
-        // frx, fry, frz, fax, fay, karaoke start, karaoke type,
-        // and on every line break
-        if (i > 0 && (last->font != info->font ||
-                    last->face_index != info->face_index ||
+        if (!info->drawing_text.str && !info->skip) {
+            // get font face and glyph index
+            ass_font_get_index(render_priv->fontselect, info->font,
+                    info->symbol, &info->face_index, &info->glyph_index);
+        }
+        if (i > 0) {
+            GlyphInfo *last = glyphs + i - 1;
+            if ((last->font != info->font ||
+                    (!info->skip &&
+                        last->face_index != info->face_index) ||
                     last->script != info->script ||
-                    last->font_size != info->font_size ||
-                    last->c[0] != info->c[0] ||
-                    last->c[1] != info->c[1] ||
-                    last->c[2] != info->c[2] ||
-                    last->c[3] != info->c[3] ||
-                    last->be != info->be ||
-                    last->blur != info->blur ||
-                    last->shadow_x != info->shadow_x ||
-                    last->shadow_y != info->shadow_y ||
-                    last->frx != info->frx ||
-                    last->fry != info->fry ||
-                    last->frz != info->frz ||
-                    last->fax != info->fax ||
-                    last->fay != info->fay ||
-                    last->scale_x != info->scale_x ||
-                    last->scale_y != info->scale_y ||
-                    last->border_style != info->border_style ||
-                    last->border_x != info->border_x ||
-                    last->border_y != info->border_y ||
-                    last->hspacing != info->hspacing ||
-                    last->italic != info->italic ||
-                    last->bold != info->bold ||
+                    info->starts_new_run ||
+                    (!shaper->whole_text_layout && info->hspacing) ||
                     last->flags != info->flags))
-            shape_run++;
+                shape_run++;
+            else if (info->skip)
+                info->face_index = last->face_index;
+        }
         info->shape_run_id = shape_run;
     }
 }
@@ -769,6 +910,10 @@ void ass_shaper_find_runs(ASS_Shaper *shaper, ASS_Renderer *render_priv,
 void ass_shaper_set_base_direction(ASS_Shaper *shaper, FriBidiParType dir)
 {
     shaper->base_direction = dir;
+
+    if (shaper->whole_text_layout != WHOLE_TEXT_LAYOUT_EXPLICIT)
+        shaper->whole_text_layout = dir == FRIBIDI_PAR_ON ?
+            WHOLE_TEXT_LAYOUT_IMPLICIT : WHOLE_TEXT_LAYOUT_OFF;
 }
 
 /**
@@ -778,7 +923,6 @@ void ass_shaper_set_base_direction(ASS_Shaper *shaper, FriBidiParType dir)
  */
 void ass_shaper_set_language(ASS_Shaper *shaper, const char *code)
 {
-#ifdef CONFIG_HARFBUZZ
     hb_language_t lang;
 
     if (code)
@@ -787,7 +931,6 @@ void ass_shaper_set_language(ASS_Shaper *shaper, const char *code)
         lang = HB_LANGUAGE_INVALID;
 
     shaper->language = lang;
-#endif
 }
 
 /**
@@ -798,94 +941,114 @@ void ass_shaper_set_level(ASS_Shaper *shaper, ASS_ShapingLevel level)
     shaper->shaping_level = level;
 }
 
-/**
-  * \brief Remove all zero-width invisible characters from the text.
-  * \param text_info text
-  */
-static void ass_shaper_skip_characters(TextInfo *text_info)
+#ifdef USE_FRIBIDI_EX_API
+void ass_shaper_set_bidi_brackets(ASS_Shaper *shaper, bool match_brackets)
 {
-    int i;
-    GlyphInfo *glyphs = text_info->glyphs;
+    shaper->bidi_brackets = match_brackets;
+}
+#endif
 
-    for (i = 0; i < text_info->length; i++) {
-        // Skip direction override control characters
-        if ((glyphs[i].symbol <= 0x202e && glyphs[i].symbol >= 0x202a)
-                || (glyphs[i].symbol <= 0x200f && glyphs[i].symbol >= 0x200b)
-                || (glyphs[i].symbol <= 0x2063 && glyphs[i].symbol >= 0x2060)
-                || glyphs[i].symbol == 0xfeff
-                || glyphs[i].symbol == 0x00ad
-                || glyphs[i].symbol == 0x034f) {
-            glyphs[i].symbol = 0;
-            glyphs[i].skip++;
-        }
-    }
+void ass_shaper_set_whole_text_layout(ASS_Shaper *shaper, bool enable)
+{
+    shaper->whole_text_layout = enable ?
+        WHOLE_TEXT_LAYOUT_EXPLICIT :
+        shaper->base_direction == FRIBIDI_PAR_ON ?
+            WHOLE_TEXT_LAYOUT_IMPLICIT : WHOLE_TEXT_LAYOUT_OFF;
 }
 
 /**
  * \brief Shape an event's text. Calculates directional runs and shapes them.
  * \param text_info event's text
+ * \return success, when 0
  */
-void ass_shaper_shape(ASS_Shaper *shaper, TextInfo *text_info)
+bool ass_shaper_shape(ASS_Shaper *shaper, TextInfo *text_info)
 {
-    int i, last_break;
-    FriBidiParType dir;
+    int i, ret, last_break;
+    FriBidiParType dir, *pdir;
     GlyphInfo *glyphs = text_info->glyphs;
+    shaper->event_text = text_info->event_text;
 
-    check_allocations(shaper, text_info->length);
+    int n_pars = 1;
+    for (i = 0; i < text_info->length - 1; i++)
+        if (glyphs[i].symbol == '\n')
+            n_pars++;
 
-    // Get bidi character types and embedding levels
-    last_break = 0;
-    for (i = 0; i < text_info->length; i++) {
+    if (!check_allocations(shaper, text_info->length, n_pars))
+        return false;
+
+    for (i = 0; i < text_info->length; i++)
         shaper->event_text[i] = glyphs[i].symbol;
+
+    fribidi_get_bidi_types(shaper->event_text,
+            text_info->length, shaper->ctypes);
+
+#ifdef USE_FRIBIDI_EX_API
+    if (shaper->bidi_brackets) {
+        fribidi_get_bracket_types(shaper->event_text,
+                text_info->length, shaper->ctypes, shaper->btypes);
+    }
+#endif
+
+    // Get bidi embedding levels
+    last_break = 0;
+    pdir = shaper->pbase_dir;
+    for (i = 0; i < text_info->length; i++) {
         // embedding levels should be calculated paragraph by paragraph
-        if (glyphs[i].symbol == '\n' || i == text_info->length - 1) {
+        if (glyphs[i].symbol == '\n' || i == text_info->length - 1 ||
+                (!shaper->whole_text_layout &&
+                    (glyphs[i + 1].starts_new_run || glyphs[i].hspacing))) {
             dir = shaper->base_direction;
-            fribidi_get_bidi_types(shaper->event_text + last_break,
-                    i - last_break + 1, shaper->ctypes + last_break);
-            fribidi_get_par_embedding_levels(shaper->ctypes + last_break,
+#ifdef USE_FRIBIDI_EX_API
+            FriBidiBracketType *btypes = NULL;
+            if (shaper->bidi_brackets)
+                btypes = shaper->btypes + last_break;
+            ret = fribidi_get_par_embedding_levels_ex(
+                    shaper->ctypes + last_break, btypes,
                     i - last_break + 1, &dir, shaper->emblevels + last_break);
+#else
+            ret = fribidi_get_par_embedding_levels(shaper->ctypes + last_break,
+                    i - last_break + 1, &dir, shaper->emblevels + last_break);
+#endif
+            if (ret == 0)
+                return false;
             last_break = i + 1;
+            if (shaper->whole_text_layout)
+                *pdir++ = dir;
         }
     }
 
-    // add embedding levels to shape runs for final runs
-    for (i = 0; i < text_info->length; i++) {
-        glyphs[i].shape_run_id += shaper->emblevels[i];
-    }
-
-#ifdef CONFIG_HARFBUZZ
     switch (shaper->shaping_level) {
     case ASS_SHAPING_SIMPLE:
         shape_fribidi(shaper, glyphs, text_info->length);
-        ass_shaper_skip_characters(text_info);
-        break;
+        return true;
     case ASS_SHAPING_COMPLEX:
-        shape_harfbuzz(shaper, glyphs, text_info->length);
-        break;
+    default:
+        return shape_harfbuzz(shaper, glyphs, text_info->length);
     }
-#else
-        shape_fribidi(shaper, glyphs, text_info->length);
-        ass_shaper_skip_characters(text_info);
-#endif
 }
 
 /**
- * \brief Create a new shaper instance and preallocate data structures
- * \param prealloc preallocation size
+ * \brief Create a new shaper instance
  */
-ASS_Shaper *ass_shaper_new(size_t prealloc)
+ASS_Shaper *ass_shaper_new(Cache *metrics_cache)
 {
+    assert(metrics_cache);
+
     ASS_Shaper *shaper = calloc(sizeof(*shaper), 1);
+    if (!shaper)
+        return NULL;
 
     shaper->base_direction = FRIBIDI_PAR_ON;
-    check_allocations(shaper, prealloc);
 
-#ifdef CONFIG_HARFBUZZ
-    init_features(shaper);
-    shaper->metrics_cache = ass_glyph_metrics_cache_create();
-#endif
+    if (!init_features(shaper))
+        goto error;
+    shaper->metrics_cache = metrics_cache;
 
     return shaper;
+
+error:
+    ass_shaper_free(shaper);
+    return NULL;
 }
 
 
@@ -910,45 +1073,60 @@ void ass_shaper_cleanup(ASS_Shaper *shaper, TextInfo *text_info)
 
 /**
  * \brief Calculate reorder map to render glyphs in visual order
+ * \param shaper shaper instance
+ * \param text_info text to be reordered
+ * \return map of reordered characters, or NULL
  */
 FriBidiStrIndex *ass_shaper_reorder(ASS_Shaper *shaper, TextInfo *text_info)
 {
-    int i;
+    int i, ret;
 
     // Initialize reorder map
     for (i = 0; i < text_info->length; i++)
         shaper->cmap[i] = i;
 
-    // Create reorder map line-by-line
-    for (i = 0; i < text_info->n_lines; i++) {
-        LineInfo *line = text_info->lines + i;
-        FriBidiParType dir = FRIBIDI_PAR_ON;
+    // Create reorder map line-by-line or run-by-run
+    int last_break = 0;
+    FriBidiParType *pdir = shaper->whole_text_layout ?
+        shaper->pbase_dir : &shaper->base_direction;
+    GlyphInfo *glyphs = text_info->glyphs;
+    for (i = 0; i < text_info->length; i++) {
+        if (i == text_info->length - 1 || glyphs[i + 1].linebreak ||
+                (!shaper->whole_text_layout &&
+                    (glyphs[i + 1].starts_new_run || glyphs[i].hspacing))) {
+            ret = fribidi_reorder_line(0,
+                    shaper->ctypes, i - last_break + 1, last_break, *pdir,
+                    shaper->emblevels, NULL,
+                    shaper->cmap);
+            if (ret == 0)
+                return NULL;
 
-        fribidi_reorder_line(0,
-                shaper->ctypes + line->offset, line->len, 0, dir,
-                shaper->emblevels + line->offset, NULL,
-                shaper->cmap + line->offset);
+            last_break = i + 1;
+            if (shaper->whole_text_layout && glyphs[i].symbol == '\n')
+                pdir++;
+        }
     }
 
     return shaper->cmap;
 }
 
+FriBidiStrIndex *ass_shaper_get_reorder_map(ASS_Shaper *shaper)
+{
+    return shaper->cmap;
+}
+
 /**
- * \brief Resolve a Windows font charset number to a suitable
- * base direction. 177 and 178 are Hebrew and Arabic respectively, and
- * they map to RTL. Everything else maps to LTR for compatibility
- * reasons. The special value -1, which is not a legal Windows font charset
- * number, can be used for autodetection.
+ * \brief Resolve a Windows font charset number to a suitable base
+ * direction. Generally, use LTR for compatibility with VSFilter. The
+ * special value -1, which is not a legal Windows font charset number,
+ * can be used for autodetection.
  * \param enc Windows font encoding
  */
-FriBidiParType resolve_base_direction(int enc)
+FriBidiParType ass_resolve_base_direction(int enc)
 {
     switch (enc) {
         case -1:
             return FRIBIDI_PAR_ON;
-        case 177:
-        case 178:
-            return FRIBIDI_PAR_RTL;
         default:
             return FRIBIDI_PAR_LTR;
     }
