@@ -17,53 +17,75 @@
  */
 
 #include "config.h"
+#include "ass_compat.h"
 
 #include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdint.h>
 #include <inttypes.h>
-#include <strings.h>
-#include <limits.h>
 
 #include "ass_library.h"
 #include "ass.h"
 #include "ass_utils.h"
+#include "ass_string.h"
 
-#if (defined(__i386__) || defined(__x86_64__)) && CONFIG_ASM
+#if CONFIG_ASM && ARCH_X86
 
 #include "x86/cpuid.h"
 
-int has_sse2(void)
+void ass_cpu_capabilities(bool *sse2, bool *avx2)
 {
-    uint32_t eax = 1, ebx, ecx, edx;
-    ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    return (edx >> 26) & 0x1;
-}
+    *sse2 = false;
+    *avx2 = false;
 
-int has_avx(void)
-{
-    uint32_t eax = 1, ebx, ecx, edx;
-    ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    if(!(ecx & (1 << 27))) // not OSXSAVE
-        return 0;
-    uint32_t misc = ecx;
-    eax = 0;
-    ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    return (ecx & 0x6) == 0x6 ? (misc >> 28) & 0x1 : 0; // check high bits are relevant, then AVX support
-}
+    if (!ass_has_cpuid())
+        return;
 
-int has_avx2(void)
-{
-    uint32_t eax = 7, ebx, ecx, edx;
+    uint32_t eax = 0, ebx, ecx, edx;
     ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    return (ebx >> 5) & has_avx();
-}
+    uint32_t max_leaf = eax;
+    bool avx = false;
 
+    if (max_leaf >= 1) {
+        eax = 1;
+        ass_get_cpuid(&eax, &ebx, &ecx, &edx);
+        if (edx & (1 << 26))  // SSE2
+            *sse2 = true;
+
+        if (ecx & (1 << 27) &&  // OSXSAVE
+            ecx & (1 << 28)) {  // AVX
+            uint32_t xcr0l, xcr0h;
+            ass_get_xgetbv(0, &xcr0l, &xcr0h);
+            if (xcr0l & (1 << 1) &&  // XSAVE for XMM
+                xcr0l & (1 << 2))    // XSAVE for YMM
+                    avx = true;
+        }
+    }
+
+    if (max_leaf >= 7) {
+        eax = 7;
+        ass_get_cpuid(&eax, &ebx, &ecx, &edx);
+        if (avx && ebx & (1 << 5))  // AVX2
+            *avx2 = true;
+    }
+}
 #endif // ASM
 
+// Fallbacks
+#ifndef HAVE_STRDUP
+char *ass_strdup_fallback(const char *str)
+{
+    size_t len    = strlen(str) + 1;
+    char *new_str = malloc(len);
+    if (new_str)
+        memcpy(new_str, str, len);
+    return new_str;
+}
+#endif
+
 #ifndef HAVE_STRNDUP
-char *ass_strndup(const char *s, size_t n)
+char *ass_strndup_fallback(const char *s, size_t n)
 {
     char *end = memchr(s, 0, n);
     size_t len = end ? end - s : n;
@@ -76,12 +98,13 @@ char *ass_strndup(const char *s, size_t n)
 }
 #endif
 
-void *ass_aligned_alloc(size_t alignment, size_t size)
+void *ass_aligned_alloc(size_t alignment, size_t size, bool zero)
 {
     assert(!(alignment & (alignment - 1))); // alignment must be power of 2
     if (size >= SIZE_MAX - alignment - sizeof(void *))
         return NULL;
-    char *allocation = malloc(size + sizeof(void *) + alignment - 1);
+    char *allocation = zero ? calloc(size + sizeof(void *) + alignment - 1, 1)
+                            : malloc(size + sizeof(void *) + alignment - 1);
     if (!allocation)
         return NULL;
     char *ptr = allocation + sizeof(void *);
@@ -133,130 +156,7 @@ void *ass_try_realloc_array(void *ptr, size_t nmemb, size_t size)
     }
 }
 
-void skip_spaces(char **str)
-{
-    char *p = *str;
-    while ((*p == ' ') || (*p == '\t'))
-        ++p;
-    *str = p;
-}
-
-void rskip_spaces(char **str, char *limit)
-{
-    char *p = *str;
-    while ((p > limit) && ((p[-1] == ' ') || (p[-1] == '\t')))
-        --p;
-    *str = p;
-}
-
-int mystrtoi(char **p, int *res)
-{
-    char *start = *p;
-    double temp_res = ass_strtod(*p, p);
-    *res = (int) (temp_res + (temp_res > 0 ? 0.5 : -0.5));
-    return *p != start;
-}
-
-int mystrtoll(char **p, long long *res)
-{
-    char *start = *p;
-    double temp_res = ass_strtod(*p, p);
-    *res = (long long) (temp_res + (temp_res > 0 ? 0.5 : -0.5));
-    return *p != start;
-}
-
-int mystrtou32(char **p, int base, uint32_t *res)
-{
-    char *start = *p;
-    *res = strtoll(*p, p, base);
-    return *p != start;
-}
-
-int mystrtod(char **p, double *res)
-{
-    char *start = *p;
-    *res = ass_strtod(*p, p);
-    return *p != start;
-}
-
-uint32_t string2color(ASS_Library *library, char *p, int hex)
-{
-    uint32_t color = 0;
-    int base = hex ? 16 : 10;
-
-    if (*p == '&')
-        while (*p == '&')
-            ++p;
-    else
-        ass_msg(library, MSGL_DBG2, "suspicious color format: \"%s\"\n", p);
-
-    if (*p == 'H' || *p == 'h') {
-        ++p;
-        mystrtou32(&p, 16, &color);
-    } else
-        mystrtou32(&p, base, &color);
-
-    while (*p == '&' || *p == 'H')
-        ++p;
-
-    unsigned char *tmp = (unsigned char *) (&color);
-    unsigned char b;
-    b = tmp[0];
-    tmp[0] = tmp[3];
-    tmp[3] = b;
-    b = tmp[1];
-    tmp[1] = tmp[2];
-    tmp[2] = b;
-
-    return color;
-}
-
-// Return a boolean value for a string
-char parse_bool(char *str)
-{
-    skip_spaces(&str);
-    return !strncasecmp(str, "yes", 3) || strtol(str, NULL, 10) > 0;
-}
-
-int parse_ycbcr_matrix(char *str)
-{
-    skip_spaces(&str);
-    if (*str == '\0')
-        return YCBCR_DEFAULT;
-
-    char *end = str + strlen(str);
-    rskip_spaces(&end, str);
-
-    // Trim a local copy of the input that we know is safe to
-    // modify. The buffer is larger than any valid string + NUL,
-    // so we can simply chop off the rest of the input.
-    char buffer[16];
-    size_t n = FFMIN(end - str, sizeof buffer - 1);
-    memcpy(buffer, str, n);
-    buffer[n] = '\0';
-
-    if (!strcasecmp(buffer, "none"))
-        return YCBCR_NONE;
-    if (!strcasecmp(buffer, "tv.601"))
-        return YCBCR_BT601_TV;
-    if (!strcasecmp(buffer, "pc.601"))
-        return YCBCR_BT601_PC;
-    if (!strcasecmp(buffer, "tv.709"))
-        return YCBCR_BT709_TV;
-    if (!strcasecmp(buffer, "pc.709"))
-        return YCBCR_BT709_PC;
-    if (!strcasecmp(buffer, "tv.240m"))
-        return YCBCR_SMPTE240M_TV;
-    if (!strcasecmp(buffer, "pc.240m"))
-        return YCBCR_SMPTE240M_PC;
-    if (!strcasecmp(buffer, "tv.fcc"))
-        return YCBCR_FCC_TV;
-    if (!strcasecmp(buffer, "pc.fcc"))
-        return YCBCR_FCC_PC;
-    return YCBCR_UNKNOWN;
-}
-
-void ass_msg(ASS_Library *priv, int lvl, char *fmt, ...)
+void ass_msg(ASS_Library *priv, int lvl, const char *fmt, ...)
 {
     va_list va;
     va_start(va, fmt);
@@ -327,14 +227,72 @@ unsigned ass_utf8_put_char(char *dest, uint32_t ch)
 }
 
 /**
- * \brief find style by name
+ * \brief Parse UTF-16 and return the code point of the sequence starting at src.
+ * \param src pointer to a pointer to the start of the UTF-16 data
+ *            (will be set to the start of the next code point)
+ * \return the code point
+ */
+static uint32_t ass_read_utf16be(uint8_t **src, size_t bytes)
+{
+    if (bytes < 2)
+        goto too_short;
+
+    uint32_t cp = ((*src)[0] << 8) | (*src)[1];
+    *src += 2;
+    bytes -= 2;
+
+    if (cp >= 0xD800 && cp <= 0xDBFF) {
+        if (bytes < 2)
+            goto too_short;
+
+        uint32_t cp2 = ((*src)[0] << 8) | (*src)[1];
+
+        if (cp2 < 0xDC00 || cp2 > 0xDFFF)
+            return 0xFFFD;
+
+        *src += 2;
+
+        cp = 0x10000 + ((cp - 0xD800) << 10) + (cp2 - 0xDC00);
+    }
+
+    if (cp >= 0xDC00 && cp <= 0xDFFF)
+        return 0xFFFD;
+
+    return cp;
+
+too_short:
+    *src += bytes;
+    return 0xFFFD;
+}
+
+void ass_utf16be_to_utf8(char *dst, size_t dst_size, uint8_t *src, size_t src_size)
+{
+    uint8_t *end = src + src_size;
+
+    if (!dst_size)
+        return;
+
+    while (src < end) {
+        uint32_t cp = ass_read_utf16be(&src, end - src);
+        if (dst_size < 5)
+            break;
+        unsigned s = ass_utf8_put_char(dst, cp);
+        dst += s;
+        dst_size -= s;
+    }
+
+    *dst = '\0';
+}
+
+/**
+ * \brief find style by name the common way (\r matches differently)
  * \param track track
  * \param name style name
  * \return index in track->styles
  * Returns 0 if no styles found => expects at least 1 style.
  * Parsing code always adds "Default" style in the beginning.
  */
-int lookup_style(ASS_Track *track, char *name)
+int ass_lookup_style(ASS_Track *track, char *name)
 {
     int i;
     // '*' seem to mean literally nothing;
@@ -343,7 +301,7 @@ int lookup_style(ASS_Track *track, char *name)
         ++name;
     // VSFilter then normalizes the case of "Default"
     // (only in contexts where this function is called)
-    if (strcasecmp(name, "Default") == 0)
+    if (ass_strcasecmp(name, "Default") == 0)
         name = "Default";
     for (i = track->n_styles - 1; i >= 0; --i) {
         if (strcmp(track->styles[i].Name, name) == 0)
@@ -355,70 +313,3 @@ int lookup_style(ASS_Track *track, char *name)
             track, name, track->styles[i].Name);
     return i;
 }
-
-/**
- * \brief find style by name as in \r
- * \param track track
- * \param name style name
- * \param len style name length
- * \return style in track->styles
- * Returns NULL if no style has the given name.
- */
-ASS_Style *lookup_style_strict(ASS_Track *track, char *name, size_t len)
-{
-    int i;
-    for (i = track->n_styles - 1; i >= 0; --i) {
-        if (strncmp(track->styles[i].Name, name, len) == 0 &&
-            track->styles[i].Name[len] == '\0')
-            return track->styles + i;
-    }
-    ass_msg(track->library, MSGL_WARN,
-            "[%p]: Warning: no style named '%.*s' found",
-            track, (int) len, name);
-    return NULL;
-}
-
-#ifdef CONFIG_ENCA
-void *ass_guess_buffer_cp(ASS_Library *library, unsigned char *buffer,
-                          int buflen, char *preferred_language,
-                          char *fallback)
-{
-    const char **languages;
-    size_t langcnt;
-    EncaAnalyser analyser;
-    EncaEncoding encoding;
-    char *detected_sub_cp = NULL;
-    int i;
-
-    languages = enca_get_languages(&langcnt);
-    ass_msg(library, MSGL_V, "ENCA supported languages");
-    for (i = 0; i < langcnt; i++) {
-        ass_msg(library, MSGL_V, "lang %s", languages[i]);
-    }
-
-    for (i = 0; i < langcnt; i++) {
-        const char *tmp;
-
-        if (strcasecmp(languages[i], preferred_language) != 0)
-            continue;
-        analyser = enca_analyser_alloc(languages[i]);
-        encoding = enca_analyse_const(analyser, buffer, buflen);
-        tmp = enca_charset_name(encoding.charset, ENCA_NAME_STYLE_ICONV);
-        if (tmp && encoding.charset != ENCA_CS_UNKNOWN) {
-            detected_sub_cp = strdup(tmp);
-            ass_msg(library, MSGL_INFO, "ENCA detected charset: %s", tmp);
-        }
-        enca_analyser_free(analyser);
-    }
-
-    free(languages);
-
-    if (!detected_sub_cp) {
-        detected_sub_cp = strdup(fallback);
-        ass_msg(library, MSGL_INFO,
-               "ENCA detection failed: fallback to %s", fallback);
-    }
-
-    return detected_sub_cp;
-}
-#endif
